import datetime
import json
from typing import Annotated, Optional
from urllib.parse import urlencode

import boto3
import requests
from pydantic import BaseModel, Field


class Credentials(BaseModel):
    access_key_id: Annotated[str, Field(alias="AccessKeyId")]
    secret_access_key: Annotated[str, Field(alias="SecretAccessKey")]
    session_token: Annotated[str, Field(alias="SessionToken")]
    expiration: Annotated[datetime.datetime, Field(alias="Expiration")]


def assume_role_with_web_identity(
    token: str,
    role_arn: str,
    role_session_name: str,
    profile_name: Optional[str] = None,
) -> Credentials:
    """
    Obtain temporary AWS credentials for a given role from a web identity token.

    Arguments:
        token: Web identity token used to authenticate to the role via AssumeRoleWithWebIdentity
        role_arn: ARN of Console user Role to assume
        role_session_name: name of authentication session which appears in AWS console
        profile_name: explicit profile from ~/.aws/config to use. If omitted, the default profile
            is used.

    Returns:
        temporary credentials

    """
    sts_connection = boto3.Session(profile_name=profile_name).client("sts")

    assumed_role_object = sts_connection.assume_role_with_web_identity(
        RoleArn=role_arn,
        RoleSessionName=role_session_name,
        WebIdentityToken=token,
    )

    return Credentials.model_validate(assumed_role_object["Credentials"])


def get_console_url(
    credentials: Credentials,
    issuer: str,
    *,
    region: str = "eu-west-2",
    session_duration_seconds: Optional[int] = 3600,
) -> str:
    """
    Construct an AWS console sign in URL.

    Arguments:
        credentials: AWS credentials of role or user to sign in to console
        issuer: arbitrary string representing the issuer of this sign in token
        region: initial region for AWS console
        session_duration_seconds: maximum session duration before re-authentication is required

    Returns:
        A sign in URL.

    """
    r = requests.get(
        _make_url(
            f"https://{region}.signin.aws.amazon.com/federation",
            {
                "Action": "getSigninToken",
                "SessionDuration": f"{session_duration_seconds}",
                "Session": json.dumps(
                    {
                        "sessionId": credentials.access_key_id,
                        "sessionKey": credentials.secret_access_key,
                        "sessionToken": credentials.session_token,
                    }
                ),
            },
        )
    )
    r.raise_for_status()
    signin_token = r.json()["SigninToken"]

    return _make_url(
        f"https://{region}.signin.aws.amazon.com/federation",
        {
            "Action": "login",
            "Issuer": issuer,
            "Destination": "https://console.aws.amazon.com/",
            "SigninToken": signin_token,
        },
    )


def _make_url(base_url: str, query: dict[str, str]):
    return f"{base_url}?{urlencode(query)}"
