import urllib.parse
from typing import Optional

import requests
from google import auth
from google.auth import impersonated_credentials
from google.auth.credentials import Credentials
from google.auth.transport.requests import Request
from pydantic import BaseModel


class UserInfo(BaseModel):
    id: str
    email: str


def get_default_credentials(quota_project_id: Optional[str] = None) -> Credentials:
    """
    Retrieve the application default Google access credentials from the environment.

    Arguments:
        quota_project_id: optional project id to use as the "quota project" for Google API calls

    Returns:
        Google credentials

    """
    credentials, _ = auth.default(quota_project_id=quota_project_id)
    return credentials


def get_user_info(credentials: Credentials) -> UserInfo:
    """
    Fetch information on a user given access token credentials.

    Arguments:
        credentials: credentials used to fetch user info

    Returns:
        information on the user.

    """
    credentials.refresh(Request())
    user_info_url = (
        "https://www.googleapis.com/oauth2/v1/userinfo?"
        f"access_token={urllib.parse.quote_plus(credentials.token)}"
    )
    r = requests.get(user_info_url)
    r.raise_for_status()
    return UserInfo.model_validate_json(r.content)


def get_id_token(credentials: Credentials, sa_email: str, audience: str) -> str:
    """
    Create an OpenID identity token for a service account with a specified audience.

    Arguments:
        credentials: credentials used to impersonate the service account
        sa_email: email address identifier of service account
        audience: string to include as the "aud" claim of the identity token

    Returns:
        An openID JWT id token.

    """
    sa_credentials = impersonated_credentials.Credentials(
        source_credentials=credentials,
        target_principal=sa_email,
        target_scopes=[],
        lifetime=30,
    )
    sa_id_token_credentials = impersonated_credentials.IDTokenCredentials(
        sa_credentials,
        target_audience=audience,
    )
    sa_id_token_credentials.refresh(Request())
    return sa_id_token_credentials.token
